
/*---------------------------------------------------------------------------
:  This file contains the main program for a distributed database site and
:  the initialization routine for a site.
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
:  Header files
---------------------------------------------------------------------------*/

#  include  <stdio.h>
#  include  "comm.h"
#  include  "simtypes.h"
#  include  "extdata.h"
#  include  "direct.h"

/*---------------------------------------------------------------------------
:  Global variable definition
---------------------------------------------------------------------------*/

   int                   num_sites;
   int			 num_db_items;
   session_vector_type   session_vector;
   session_vector_type   old_session_vector;
   actual_snum_type      actual_session_num;
   database_type         database;
   directory_type 	 directories; 
   fail_locks_type       fail_locks;
   man_subfunction_type  failure_schedule;
   /*msg_type		 *last_update=(msg_type *)NULL;*/
   msg_type		 *last_update; 
   int			 num_pseudo = 0;
   int			 num_copier = 0;

/*---------------------------------------------------------------------------
:  This is the main routine for the program which functions as a site
:  in a distributed database.  The site expects two arguments:  the first
:  is the site ID and the second is the total number of database sites. 
---------------------------------------------------------------------------*/

main ( argc , argv )

   int  argc;
   char *argv[];
{

/*---------------------------------------------------------------------------
:  Local variable definition
---------------------------------------------------------------------------*/

#  define  SITE_ID    argv[1]
#  define  NUM_SITES  argv[2]
#  define  NUM_ITEMS  argv[3]
   int log_site_info() , initialize_site() , process_control_msg() ,
       process_xact_msg() , process_managing_msg();
   site_state_type get_site_state() , set_site_state();
   int rcv_status;
   msg_type rcv_message;

   last_update = (msg_type *) malloc (sizeof(msg_type));
/*---------------------------------------------------------------------------
:  Start of code - initialize the site and loop to receive and process messages
---------------------------------------------------------------------------*/

   if (initialize_site(SITE_ID , NUM_SITES, NUM_ITEMS) != ERROR) {

      while (get_site_state (MySID) != SITE_TERMINATING) {
         rcv_status = receive_message (&rcv_message);

	 if (rcv_status != ERROR)
	    switch (rcv_message.type) {
	       case control :
		    start_timing ("process_control_msg");
		    process_control_msg (&rcv_message);
		    stop_timing ();
		    break;

               case xact :
		    start_timing ("process_xact_msg"); 
		    process_xact_msg (&rcv_message);
		    stop_timing (); 
		    break;

               case managing :
		    start_timing ("process_managing_msg");
		    process_managing_msg (&rcv_message);
		    stop_timing ();
		    break;

               default :
		    log_site_info (NO_DUMP , "*** undefined msg type ***");
		    break;
	    } 
	 else
	    set_site_state (MySID , SITE_TERMINATING);
      }   /* end while */

/*---------------------------------------------------------------------------
:     Shut down this site 
---------------------------------------------------------------------------*/

      log_site_info (NO_DUMP , "site stopping");
      ExitSite ();
   }

   exit (0);
}

/*---------------------------------------------------------------------------
:  This procedure performs initialization for a database site
---------------------------------------------------------------------------*/

initialize_site ( site_c , num_sites_c , num_items_c)

   char *site_c , *num_sites_c, *num_items_c;
{

/*---------------------------------------------------------------------------
:  Local variable definition
---------------------------------------------------------------------------*/

   int initialize_ctl_info() , initialize_database();
   int init_status = OK;
   char comment[40];

/*---------------------------------------------------------------------------
:  Start of code - indicate start up and set number of sites
---------------------------------------------------------------------------*/

   log_site_info (NO_DUMP , "###########################################");
   sprintf (comment , "site %s started up (total # sites = %s)" , 
	    site_c , num_sites_c);
   log_site_info (NO_DUMP , comment);
   num_sites = atoi (num_sites_c);
   num_db_items = atoi (num_items_c);

/*---------------------------------------------------------------------------
:  Initialize communications for this site
---------------------------------------------------------------------------*/

   if ((init_status = Setup (atoi (site_c))) != ERROR) {

/*---------------------------------------------------------------------------
:     Initialize control and database information
---------------------------------------------------------------------------*/

      initialize_ctl_info();
      initialize_database();
/***********/
      request_directory() ; 
      receive_directory() ; 
/***********/
   }
   else
      log_site_info (NO_DUMP , "*** error initializing communications ***");

   return (init_status);
}
