
/*---------------------------------------------------------------------------
:  This file contains the main program for the simulation manager and a
:  simulation initialization routine.
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
:  Header files
---------------------------------------------------------------------------*/

#  include  <stdio.h>
#  include  "comm.h"
#  include  "simtypes.h"
#  include  "mgr.h"
#  include  "direct.h"

/*---------------------------------------------------------------------------
:  Global variable definition
---------------------------------------------------------------------------*/

   int                   num_sites;
   actual_snum_type      actual_session_num;
   session_vector_type   session_vector;
   session_vector_type   old_session_vector;
   database_type         database;
   directory_type	 directories ; 
   int 			 replic_num ; 
   int 			 min_copies; 
   fail_locks_type       fail_locks;
   man_subfunction_type  failure_schedule;
   msg_type		 *last_update=(msg_type *)NULL;

   int                   xact_num , abort_num , copier_num [MAX_NUM_SITES] ,
			 num_f_locks [MAX_NUM_SITES] , 
			 pseudo_num [MAX_NUM_SITES] ,
			 num_copier_clr_fl [MAX_NUM_SITES];

/*---------------------------------------------------------------------------
:  This is the main routine for the program which manages the distributed
:  database simulation.
---------------------------------------------------------------------------*/

main ()
{

/*---------------------------------------------------------------------------
:  Local variable definition
---------------------------------------------------------------------------*/

   int  initialize_manager() , execute_mgr_command();
   boolean simulation_running = TRUE , try_execution;
   command_string  in_command;
   command_type    command = HELP;
   site_id_type    site;
   site_id_type    subject_site;

/*---------------------------------------------------------------------------
:  Start of code - set up the simulation manager and loop to process commands
---------------------------------------------------------------------------*/

   CreatePortFile();
   if (initialize_manager() != ERROR) {

      while (simulation_running) {
	 try_execution = TRUE;

/*---------------------------------------------------------------------------
:        Get a site identifier if necessary before command execution
---------------------------------------------------------------------------*/

	 if ((command == RECOVER_SITE) || (command == FAIL_SITE) || 
	     (command == DUMP_SITE_INFO) || (command == SEND_TRANSACTION) ||
	     (command == SEND_ALLOW_RECOVERY) || (command == 
	      BUILD_OWN_TRANSACTION)) {
            printf ("    Destination site ID [%1d:%1d]:  " , FIRST_SITE_ID , 
		 	 num_sites-1);
	    scanf ("%d" , &site);

	    if ((site < FIRST_SITE_ID) || (site >= num_sites)) {
	       try_execution = FALSE;
	       printf ("*** Error:  illegal site ID ***\n");
	    }
	    
	    if (command == SEND_ALLOW_RECOVERY) {
            	   printf("    Object site ID [%1d:%1d]:  ",FIRST_SITE_ID,
			       num_sites-1);
	    	   scanf ("%d" , &subject_site);

	           if((subject_site<FIRST_SITE_ID)||(subject_site >= num_sites))
	              {
	      	         try_execution = FALSE;
	       	         printf ("*** Error:  illegal site ID ***\n");
                      }
            }
         }

	 if (try_execution) execute_mgr_command (command , site, subject_site);

/*---------------------------------------------------------------------------
:        Get the next simulation command
---------------------------------------------------------------------------*/

         if (command != STOP_SIMULATION) {
            printf ("%s" , PROMPT);
	    scanf ("%s" , in_command);
	    command = in_command[0];
	 }
         else
	    simulation_running = FALSE;

      }  /* end while */

/*---------------------------------------------------------------------------
:     Shut down the communications for this site 
---------------------------------------------------------------------------*/

      ExitSite ();
   }
   DeletePortFile();
   exit (0);
}

/*---------------------------------------------------------------------------
:  This procedure performs initialization for simulation management
---------------------------------------------------------------------------*/

initialize_manager ()
{

/*---------------------------------------------------------------------------
:  Local variable definition
---------------------------------------------------------------------------*/

#  define  SITE_STARTUP_DELAY  2
#  define  WRITE  "w"
   site_state_type set_site_snum();
   int initialize_database();
   int sys_status , init_status;
   char site_c[10] , num_db_items_c[10], num_sites_c[10]; 
   site_id_type  site_id = SIMULATION_MGR_ID , site;  
   exec_log_filename_type exec_log_filename;
   FILE *exec_log_file;
   int item, copies, random(); 
   short int mask ; 
   msg_type  rcv_message ; 
   int num_remaining, i, j ; 

/*---------------------------------------------------------------------------
:  Start of code - set up the simulation manager as a communicating site
---------------------------------------------------------------------------*/

   if ((init_status = Setup (site_id)) != ERROR) {
      initialize_database ();

/*---------------------------------------------------------------------------
:     Initialize simulation tracking variables
---------------------------------------------------------------------------*/

      xact_num = 0;
      abort_num = 0;
      for (site = FIRST_SITE_ID ; site < MAX_NUM_SITES ; site++) {
	 copier_num [site] = 0;
	 pseudo_num [site] = 0;
	 num_f_locks [site] = 0;
	 num_copier_clr_fl [site] = 0;
      }

/*---------------------------------------------------------------------------
:     Get the number of operations per user transaction
---------------------------------------------------------------------------*/

      num_xact_operations = 0;
      while ((num_xact_operations <= 0) || 
             (num_xact_operations > MAX_NUM_XACT_ACTIONS)) {
         printf ("\nEnter the number of operations for a user xact [1-%2d]:  " ,
	         MAX_NUM_XACT_ACTIONS);
         scanf ("%d" , &num_xact_operations);
      }

/*---------------------------------------------------------------------------
:     Get the number of database items to be involved in the simulation
:     (this capability of using less than the maximum defined is to allow
:      more simulation flexibility)
---------------------------------------------------------------------------*/

      num_db_items = 0;
      while ((num_db_items <= 0) || (num_db_items > MAX_NUM_ITEMS)) {
         printf ("\nEnter the number of data-items for simulation [1-%2d]:  " ,
	         MAX_NUM_ITEMS);
         scanf ("%d" , &num_db_items);
      }
      sprintf(num_db_items_c, "%d", num_db_items) ; 

/*---------------------------------------------------------------------------
:     Get the number of sites to be started up
---------------------------------------------------------------------------*/

      num_sites = 0;
      while ((num_sites < MIN_NUM_SITES) || (num_sites > MAX_NUM_SITES)) {
         printf ("\nEnter the number of sites to be started up [%1d-%1d]:  " ,
	         MIN_NUM_SITES , MAX_NUM_SITES);
         scanf ("%d" , &num_sites);
      }

/*****************************
 get number of replicated copies for each site 
*****************************/
	replic_num = 0 ; 
	while (replic_num < 1 || replic_num > num_sites ) { 
	      printf("\nEnter number of replicated copies [1-%2d]: ", 
			num_sites);
	      scanf("%d", &replic_num ) ; 
	} 
/*****************************
get minimum threshold 
*****************************/
	if (replic_num > 1 ) { 
	   min_copies = 0 ; 
	   while (min_copies < 1 || min_copies > replic_num ) { 
		 printf("\nEnter minimum number of copies maintained [1-%2d]: ",
				replic_num ) ; 
		scanf("%d", &min_copies) ; 
	   } 
	} 
/*****************************
 distribute the items to the sites
*****************************/
	for (item = 0 ; item < num_db_items ; item++) { 
	    mask = 0 ; 
	    num_remaining = num_sites ; 
	    site = (random()%num_sites); 
	    for (copies = 0 ; copies < replic_num;) { 
		 for (j = 0 , i = 0 ; i < site ; j++)  { 
		      if (!TEST_BIT(mask, j) ) 
			 i++ ; 
		 } 
		 while (TEST_BIT(mask, j) ) 
 			j++ ; 
		 if (j > num_sites ) 
		     printf("Error in random distribution\n"); 
		 SET_BIT(mask,j); 	
		 copies++ ; 
		 num_remaining-- ; 
		 if (num_remaining)
		    site = (random()% num_remaining); 
	    } 
	    SET_BIT(mask, SIMULATION_MGR_ID) ; 
	    directories[item].direct = mask ; 
	    directories[item].min_copies = min_copies ; 
	    directories[item].num_copies = replic_num ; 
	} 
      sprintf (num_sites_c , "%d" , num_sites);

/*---------------------------------------------------------------------------
:     Start up the requested number of sites
---------------------------------------------------------------------------*/

      for (site = FIRST_SITE_ID ; site < num_sites ; site++) {

/*---------------------------------------------------------------------------
:        Initialize the execution log file for the database site 
---------------------------------------------------------------------------*/

	 sprintf (site_c , "%d" , site); 
	 strcpy (exec_log_filename , EXEC_LOG_ROOT_FILENAME);
	 strcat (exec_log_filename , site_c);
	 exec_log_file = fopen (exec_log_filename , WRITE);

/*---------------------------------------------------------------------------
:        To start up a site, fork a process and have the child execute the
:        site program (note that the site program needs the site id and the
:        total number of sites as arguments).  The child uses the log file as 
:        its standard output and standard error.  The parent manager process 
:        can update the site state in its manager session vector.
---------------------------------------------------------------------------*/

	 sys_status = fork();

	 if (sys_status == 0) {
	    dup2 (fileno (exec_log_file) , fileno (stdout));
	    dup2 (fileno (exec_log_file) , fileno (stderr));

	    if (execl (DB_SITE_FILE , DB_SITE_PROGRAM , site_c , 
		       num_sites_c , num_db_items_c, NULL) == -1) {
	       printf ("\n*** Child unable to start site program ***\n");
               exit (0);
	    }
         }
	 else if (sys_status > 0) {
	    printf ("Start-up performed for site %-2d\n" , site);
	    set_site_snum (site , INITIAL_SNUM);
	 }
	 else
	    printf ("*** Unable to fork child process for site %-2d\n" , site);

/*---------------------------------------------------------------------------
:        Close the site log file and sleep to give the site time to come up
---------------------------------------------------------------------------*/

         fclose (exec_log_file);
	 sleep (SITE_STARTUP_DELAY);
      }  /* end for */
   }	  
   else
      printf ("\n*** Error initializing communications ***\n"); 

/*******************************
  wait for the directory requests and initiate a send
********************************/
   for (site = FIRST_SITE_ID; site < num_sites ; site++) { 
       if (receive_message(&rcv_message) != ERROR) { 
	  if (rcv_message.type == control && 
		rcv_message.subtype == directory_copy) 
	     send_directory(rcv_message.source) ; 
	  else { 
	     printf("** -- ERROR mgr.c ; wrong message received -- **\n") ; 
	     init_status = ERROR; 
	  } 
       } 
       else { 
	  printf("** -- ERROR mgr.c ; can't recieve message -- **\n") ; 
          init_status = ERROR; 
       } 
   } 
   return (init_status);
}
