/*---------------------------------------------------------------------------
:  This file contains procedures for the execution of manager commands
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
:  Header files
---------------------------------------------------------------------------*/

#  include  <stdio.h>
#  include  <sys/file.h>
#  include  <unistd.h>
#  include  "comm.h"
#  include  "simtypes.h"
#  include  "extdata.h"
#  include  "mgr.h"
#   include "direct.h"

/*---------------------------------------------------------------------------
:  Define global variables for tracking the simulation's progress
---------------------------------------------------------------------------*/

   extern int xact_num , abort_num , copier_num [MAX_NUM_SITES] ,
      num_f_locks [MAX_NUM_SITES] , num_copier_clr_fl [MAX_NUM_SITES] ,
      pseudo_num[MAX_NUM_SITES];

/* moved prototype here to avoid scope warnings - DM */
   site_state_type get_site_state();

/*---------------------------------------------------------------------------
:  execute_mgr_command oversees execution of a simulation command
---------------------------------------------------------------------------*/

execute_mgr_command ( command , site , subject_site )

   command_type  command;
   site_id_type  site;
   site_id_type  subject_site;
{

/*---------------------------------------------------------------------------
:  Local variable definition
---------------------------------------------------------------------------*/

/*DM   site_state_type get_site_state(); */
   int send_message();
   site_id_type  cur_site;  
   int num_to_send;
   msg_type message;
   char open_mode;
   command_string answer;
   boolean site_cleared , fail_locks_found[MAX_NUM_SITES];
   FILE *mgr_log_file;

/*---------------------------------------------------------------------------
:  Start of code - process manager command
---------------------------------------------------------------------------*/

   switch (command)  {
      case HELP :
	   printf ("\nSimulation commands:\n");
	   printf ("   %c = help\n" , HELP);
	   printf ("   %c = fail site\n" , FAIL_SITE);
	   printf ("   %c = recover site\n" , RECOVER_SITE);
           printf ("   %c = send user transaction\n" , SEND_TRANSACTION);
	   printf ("   %c = build own transaction\n" , BUILD_OWN_TRANSACTION);
	   printf ("   %c = send multiple user transactions to random sites\n" ,
		   SEND_MULTIPLE_XACTS);
           printf ("   %c = send mult. user xacts until fail-locks cleared\n" ,
		   SEND_UNTIL_CLEAR);
	   printf ("   %c = cause site to dump information\n" , DUMP_SITE_INFO);
           printf ("   %c = output current information\n" , OUTPUT_CUR_INFO);
	   printf ("   %c = output information summary\n" , SUMMARIZE_INFO);
	   printf ("   %c = check on children\n" , CHECK_CHILDREN);
	   printf ("   %c = send allow recovery\n" , SEND_ALLOW_RECOVERY);
	   printf ("   %c = stop simulation\n" , STOP_SIMULATION);
	   printf ("\n");
	   break;

      case FAIL_SITE :
	   message.source = SIMULATION_MGR_ID;
	   message.dest = site;
	   message.type = managing;
	   message.subtype = die;
           get_failure_time (&message);
	   send_message (&message);
           break;

      case RECOVER_SITE :
	   message.source = SIMULATION_MGR_ID;
	   message.dest = site;
	   message.type = managing;
	   message.subtype = revive;
	   send_message (&message);
           break;

      case SEND_TRANSACTION :
      case BUILD_OWN_TRANSACTION: 
	   update_site_info ();
	   process_xact_user_msg (&message , site, command);
           break;

      case SEND_MULTIPLE_XACTS :
	   update_site_info ();
	   printf ("\nEnter number of transactions to send:  ");
	   scanf ("%d" , &num_to_send);
	   while (num_to_send > 0) {
              cur_site = get_random_up_site (); 
	      if (cur_site != NULL_SITE_ID) {
  	         process_xact_user_msg (&message , cur_site, SEND_TRANSACTION);
	         num_to_send --;
              }
              else {
		 printf ("*** No sites are UP ***\n");
		 num_to_send = 0;
	      }
	   }
           break; 

      case SEND_UNTIL_CLEAR :
	   update_site_info ();
           site_cleared = TRUE;
	   for (cur_site = FIRST_SITE_ID ; cur_site < num_sites ; cur_site++)
	      if ((get_site_state (cur_site) == SITE_UP) && 
		  (num_f_locks[cur_site] > 0)) {
		 site_cleared = FALSE;
		 fail_locks_found[cur_site] = TRUE;
	      }
              else
		 fail_locks_found[cur_site] = FALSE;

	   while (! site_cleared) {
              cur_site = get_random_up_site (); 
	      if (cur_site != NULL_SITE_ID) {
  	         process_xact_user_msg (&message , cur_site, SEND_TRANSACTION);
	         for (cur_site = FIRST_SITE_ID ; cur_site < num_sites ;
		      cur_site++)
                    if ((fail_locks_found[cur_site]) &&
			(num_f_locks[cur_site] == 0)) {
		       site_cleared = TRUE;
		       break;
		    }
              }
              else {
		 printf ("*** No sites are UP ***\n");
		 site_cleared = TRUE;
	      }
	   }
           break;

      case DUMP_SITE_INFO :
	   message.source = SIMULATION_MGR_ID;
	   message.dest = site;
	   message.type = managing;
	   message.subtype = dump;
	   send_message (&message);
	   break;

      case OUTPUT_CUR_INFO :
	   update_site_info ();
           log_site_info (DUMP , "current information");
	   printf ("\n    Write to manager log? (%c,%c):  " , YES , NO);
	   scanf ("%s" , answer); 
	   if (answer[0] == YES) {
	      open_mode = (access (MGR_LOG_FILENAME , F_OK) == 0 ? 'a' : 'w');
	      mgr_log_file = fopen (MGR_LOG_FILENAME , &open_mode);
	      if (fork() == 0) {
	         dup2 (fileno (mgr_log_file) , fileno (stdout));
	         log_site_info (DUMP , "current information");
		 exit (0);
	      }
/*DM	      close (mgr_log_file); */
              fclose(mgr_log_file);
	   }
	   break;

      case SUMMARIZE_INFO :
	   printf ("Simulation set-up has %1d sites, " , num_sites);
	   printf ("xacts with %-2d operations, and %-2d data-items\n" ,
		   num_xact_operations , num_db_items);
	   update_site_info ();
	   log_site_info (SUMMARY , "current information");
	   break;

      case CHECK_CHILDREN :
	   system ("ps x");
	   break;

      case STOP_SIMULATION :
	   for (cur_site = FIRST_SITE_ID ; cur_site < num_sites ; cur_site++) {
	      message.source = SIMULATION_MGR_ID;
	      message.dest = cur_site;
	      message.type = managing;
	      message.subtype = stop;
	      send_message (&message);
	   }
           break;

      case SEND_ALLOW_RECOVERY :
	   send_managing_allow_recovery(site,subject_site);
	   break;

      default :
	   printf ("*** Undefined command - enter %c for help ***\n" , HELP);
           break;
            
   }  /* end switch */

   return (0);
}

/*---------------------------------------------------------------------------
:  update_site_info gets the current information on each site's status
---------------------------------------------------------------------------*/

update_site_info ()
{
   site_status_type ctl_info_io () , cur_site_status;
   site_state_type set_site_snum() , set_site_state();
   site_id_type site;

   for (site = FIRST_SITE_ID ; site < num_sites ; site++) {
      cur_site_status = ctl_info_io (READ_STATUS , site);
      set_site_snum (site , cur_site_status.session_num);
      set_site_state (site , cur_site_status.state);
   }
   return (0);
}

/*---------------------------------------------------------------------------
:  get_failure_time interacts with the user to schedule site failure
---------------------------------------------------------------------------*/

get_failure_time ( message )

   msg_type *message;
{ 
   boolean failure_scheduled = FALSE;
   command_string failure;

   while (! failure_scheduled) {
      failure_scheduled = TRUE;
      printf ("    Failure schedule (enter %c for help):  " , SET_FAIL_HELP);
      scanf ("%s" , failure);

      switch (failure[0]) {
         case SET_FAIL_HELP :
	      failure_scheduled = FALSE;
	      printf ("       %c = fail immediately\n" , SET_FAIL_NOW); 
	      printf ("       %c = fail after receiving update\n" ,
	              SET_FAIL_UPD_RECV);
              printf ("       %c = fail after sending update\n" ,
		      SET_FAIL_UPD_SEND);
              break;
              
         case SET_FAIL_NOW :
	      message->c.subfunction = FAIL_IMMEDIATELY;
	      break;

         case SET_FAIL_UPD_RECV :
	      message->c.subfunction = FAIL_AFTER_UPD_RECV;
	      break;

         case SET_FAIL_UPD_SEND :
	      message->c.subfunction = FAIL_AFTER_UPD_SEND;
	      break;

         default :
	      failure_scheduled = FALSE;
	      printf ("   *** Illegal failure specification ***\n");
	      break;
      }
   }  /* end while */

   return (0);
}

/*---------------------------------------------------------------------------
:  process_xact_user_msg builds and sends a user transaction and then waits
:  for a reply.  If the transaction has committed then update the manager's
:  version of the database.
---------------------------------------------------------------------------*/

process_xact_user_msg ( message , dest_site , command)

   msg_type *message;
   site_id_type dest_site;
   command_type command; 
{
   msg_type *rcv_msg;
   int index;
   boolean output_info = FALSE;
   site_id_type site;

   if (get_site_state (dest_site) == SITE_UP) {
      message->source = SIMULATION_MGR_ID;
      message->dest = dest_site;
      message->type = xact;
      message->subtype = user;
      if (command == BUILD_OWN_TRANSACTION) 
	 build_own_trans(message) ; 
      else 
         build_xact_user_msg (message);
      xact_num ++;
      printf ("\nSending user transaction #%-d" , xact_num);
      log_message (message);
      send_message (message);

      rcv_msg = (msg_type *) malloc (sizeof (msg_type));
      receive_message (rcv_msg);
      
      while (rcv_msg->type != managing ) { 
	    if (rcv_msg->type == control && rcv_msg->subtype == 
					recovery_response) 
		bcopy(rcv_msg->c.ctl.f_locks, fail_locks, 
					sizeof(fail_locks_type) ) ; 
	    else if (rcv_msg->type == control && rcv_msg->subtype ==
					directory_ack ) 
		    bcopy(rcv_msg->c.ctl_direct, directories, 
				sizeof(directory_type)); 
		 else if (rcv_msg->type == control && rcv_msg->subtype ==
					clear_fail_locks) 
			 process_clear_fail_locks(rcv_msg) ; 
		      else
		         printf("   **** unexpected message **** \n") ; 
	    receive_message(rcv_msg) ; 
      } 
      if (rcv_msg->type == managing) {
	 update_site_info ();
	 if (rcv_msg->subtype == xact_committed) {
	    output_info = TRUE;
	    pseudo_num[dest_site] += rcv_msg->c.db.num_pseudo_xacts;
	    if (rcv_msg->c.db.num_copier_xacts > 0) {
	       copier_num[dest_site] += rcv_msg->c.db.num_copier_xacts;

               for (index = 0 ; index < message->c.ux.num_actions ; index++)
                  if ((message->c.ux.op[index].action == READ_ITEM) &&
                      (fail_locks_mgr (CHECK_FOR_SITE , dest_site ,
				       message->c.ux.op[index].item) ==
                      FAIL_LOCK_SET)) {

		     fail_locks_mgr (CLEAR_FOR_SITE , dest_site ,
                                     message->c.ux.op[index].item);
                     (num_copier_clr_fl[dest_site]) ++; 
		  }
            }
	    commit_db_values (rcv_msg);
         }
         else if (rcv_msg->subtype == xact_aborted) {
	    output_info = TRUE;
	    abort_num ++;
	    pseudo_num[dest_site] += rcv_msg->c.db.num_pseudo_xacts;
	    if (rcv_msg->c.db.num_copier_xacts > 0) {
	       copier_num[dest_site] += rcv_msg->c.db.num_copier_xacts;
	       for (index = 0 ; index < rcv_msg->c.db.num_items ; index ++) {
		  (num_copier_clr_fl[dest_site]) ++;
		  fail_locks_mgr (CLEAR_FOR_SITE , dest_site ,
				  rcv_msg->c.db.item[index]);
	       }
            }
	 }
	 else
            printf ("    *** unexpected message ***\n");
      }
      else
         printf ("    *** unexpected message ***\n");

      free (rcv_msg);
   }
   else
      printf ("    *** the site is not UP ***\n");

/*---------------------------------------------------------------------------
:  Output information about the xact if some work was done
---------------------------------------------------------------------------*/

   if (output_info) {
      printf ("Xact #%-4.4d , Abort #%-4.4d\n" , xact_num , abort_num);
      printf ("Site   Fail-locks   Copiers   Copier fail-locks   Pseudo xacts\n"); 
      for (site = FIRST_SITE_ID ; site < num_sites ; site++) {
	 num_f_locks[site] = count_fail_locks (site);
	 printf (" #%1d        %2.2d         %4.4d           %2.2d               %4.4d\n" ,
		 site , num_f_locks[site] , copier_num[site] , 
		 num_copier_clr_fl[site], pseudo_num[site]);
      }
   }
   return (0);
}

/*---------------------------------------------------------------------------
:  build_xact_user_msg creates a database request using random numbers
---------------------------------------------------------------------------*/

build_xact_user_msg ( message )

   msg_type *message;
{
   int index;

   message->c.ux.num_actions = (random() % num_xact_operations) + 1;

   for (index = 0 ; index < message->c.ux.num_actions ; index++) {
      message->c.ux.op[index].action = 
	 ((random() % 2) == 0 ? READ_ITEM : WRITE_ITEM);
      message->c.ux.op[index].item = (data_item_type) (random() % num_db_items);
      message->c.ux.op[index].value = (data_value_type) 
	 (message->c.ux.op[index].action == WRITE_ITEM ?
	  (random() % MAX_ITEM_VALUE) + 1 : NULL_ITEM_VALUE); 
   }
   return (0);
}

/*---------------------------------------------------------------------------
:  get_random_up_site randomly chooses a site which is currently UP
---------------------------------------------------------------------------*/

site_id_type get_random_up_site ()
{
   site_id_type site_id , up_sites[MAX_NUM_SITES];
   int num_up_sites = 0;

   for (site_id = 0 ; site_id < num_sites ; site_id++)
      if (get_site_state (site_id) == SITE_UP) {
	 up_sites[num_up_sites] = site_id;
	 num_up_sites ++;
      }
   return (num_up_sites > 0 ? 
	   up_sites [random() % num_up_sites] : NULL_SITE_ID); 
}
/* 
 * build_own_trans - let the user choose the actions and items that the
 * 		     transaction should contain.
 */
build_own_trans(message)
msg_type *message ; 
{ 	int num_act , i ;
	int  item_num ; 
	/*data_value_type w_value ; */
	int w_value ; 
	db_action_type act_type; 
	command_string in_command; 

	num_act = 0 ; 
	while (num_act <= 0 || num_act > num_xact_operations) { 
	      printf("    Enter number of actions for this transactions [1-%2d]: ",
				num_xact_operations ) ; 
	      scanf("%d", &num_act ) ; 
	} 
	message->c.ux.num_actions = num_act ; 
	for (i = 0 ; i < num_act ; i++) { 
	    act_type = NULL_ACTION ; 
	    while (act_type != READ_ITEM && act_type != WRITE_ITEM ) { 
		printf("    Enter action%d [R/W]: ", i+1) ; 
		scanf("%s", in_command) ; 
		act_type = in_command[0] ; 
	    } 
	    message->c.ux.op[i].action = act_type ; 
	    item_num = -1 ; 
	    while (item_num < 0 || item_num > num_db_items -1 ) { 
		printf("    Enter item number [0-%2d]: ", num_db_items -1) ; 
		scanf("%d", &item_num ) ; 
	    } 
	    message->c.ux.op[i].item = item_num ; 
	    if (act_type == WRITE_ITEM) { 
	       w_value = 0 ; 
	       while (w_value <= 0 || w_value > MAX_ITEM_VALUE) { 
		     printf("    Enter item value [1-%2d]: ", MAX_ITEM_VALUE) ; 
		     scanf("%d", &w_value ) ; 
	       } 
	       message->c.ux.op[i].value = w_value ; 
	    } 
	    else 
	       message->c.ux.op[i].value = NULL_ITEM_VALUE ; 
	} /* end for */
} 

