#include "comm.h"
#include "simtypes.h"
#include "extdata.h"
#include "direct.h"

/*------------------------------------------------------------------------
: get_new_copies - after a failure announce, the site should check if it 
:  		   is supposed to get new copies of any objects and issues
: 		   copier transactions for them.
-------------------------------------------------------------------------*/
get_new_copies (msg, msg2) 
msg_type *msg, *msg2; 
{ 	typedef struct { 
			site_id_type 	site; 
			data_item_type  item; 
		} new_item_type  ; 
	new_item_type new[MAX_NUM_ITEMS] ; 
	int i, new_count, index , num_requested; 
	site_id_type site; 
	data_item_type item; 
	data_value_type cur_item_value ; 
	boolean check_complete = FALSE, failure_detected, rcvd_user_xact 
				= FALSE ; 
	char comment[50]; 
	msg_type *rcv_msg, *copy_msg, *usr_msg ; 

   while (! check_complete) {

/*---------------------------------------------------------------------------
:     Build a list of data-items which are new and need to be copied 
---------------------------------------------------------------------------*/
      new_count = 0;
      for (item = 0 ; item < num_db_items; item++) { 
	     if (((directory_mgr(CHECK_FOR_COPY, MySID,item) != COPY_PRESENT ) 
			&& (TEST_BIT(msg->c.ctl_direct[item].direct,MySID)))
			|| ((fail_locks_mgr(CHECK_FOR_SITE, MySID, item) ==
			   FAIL_LOCK_SET) && ( ! TEST_BIT(
			   msg2->c.ctl.f_locks[item], MySID))) ) 
		 new[new_count++].item =item ; 
      } 

/*---------------------------------------------------------------------------
:     Attempt to match each new item with a site which has a good copy.
:     If all sites have an item locked then no more checking is necessary. 
---------------------------------------------------------------------------*/

      for (index = 0 ; index < new_count ; index++) {
	 new[index].site = NULL_SITE_ID;
	 for (site = 0 ; site < num_sites ; site++)
            if ((get_site_state (site) == SITE_UP) &&
	        (fail_locks_mgr (CHECK_FOR_SITE , site , new[index].item) ==
                 FAIL_LOCK_CLEAR) && directory_mgr(CHECK_FOR_COPY, site, 
				new[index].item) == COPY_PRESENT) {

               new[index].site = site;
	       break;   /* site found for this item => break for loop */
	    }
         if (new[index].site == NULL_SITE_ID) {
	    sprintf (comment , "no sites  have new copy for %d" , 
			new[index].item);
	    log_site_info (NO_DUMP , comment);
	    check_complete = TRUE;
	    break;   /* current item missing everywhere => skip other items */
         }
      }  /* end for */

/*---------------------------------------------------------------------------
:     Get good copies of items from the other sites if necessary
---------------------------------------------------------------------------*/

      if ((new_count > 0) && (! check_complete)) {
	 /*failure_detected = FALSE;*/
         num_requested = 0;
	 rcv_msg = (msg_type *) malloc (sizeof(msg_type));
	 copy_msg = (msg_type *) malloc (sizeof(msg_type));
	 copy_msg->source = MySID;
	 copy_msg->type = xact;
	 copy_msg->subtype = copier;

	 for (site = FIRST_SITE_ID ; site < num_sites ; site++) {
	    if (site == MySID) continue;
	    copy_msg->c.db.num_items = 0; 

/*---------------------------------------------------------------------------
:           Make a list of all items to be requested from the current site
---------------------------------------------------------------------------*/

            for (index = 0 ; index < new_count ; index++){ 
	       if (site == new[index].site) 
		  copy_msg->c.db.item[copy_msg->c.db.num_items++] =
		     new[index].item;
	    }

/*---------------------------------------------------------------------------
:           If the site has some good copies then issue the copier xact
:           and process the reply
---------------------------------------------------------------------------*/

            if (copy_msg->c.db.num_items > 0) {
               num_requested += copy_msg->c.db.num_items;
	       copy_msg->dest = site;
	       send_message (copy_msg);
		  
               receive_message (rcv_msg);
	       
/*---------------------------------------------------------------------------
:              Site responded with items so update the database and add
:              items to the list of cleared items
---------------------------------------------------------------------------*/
	       while (!(rcv_msg->type == xact && rcv_msg->subtype == 
	       				copier_update) ) { 
               
/*---------------------------------------------------------------------------
:              Other site has failed so update its state 
---------------------------------------------------------------------------*/

                  if ((rcv_msg->type == managing) &&
			   (rcv_msg->subtype == failed)) {
		     failure_detected = TRUE;
                     set_site_state (rcv_msg->source , SITE_DOWN);
	          }
/*-------------------------------------------------------------------------
: 	Received a copier transaction so process it. 
-------------------------------------------------------------------------*/
                  else if ((rcv_msg->type == xact) && (rcv_msg->subtype ==
				   copier) )
			   process_copier_xact(rcv_msg) ; 
/*-------------------------------------------------------------------------
: 	Received a user transaction so save it till later.
-------------------------------------------------------------------------*/
		       else if ((rcv_msg->type == xact) && (rcv_msg->subtype 
					   == user)) { 
			        rcvd_user_xact = TRUE ; 
			        usr_msg = (msg_type *) malloc (sizeof(msg_type));
			        bcopy(rcv_msg, usr_msg, sizeof(msg_type)); 
			    } 
			    else 
		  	        log_site_info (NO_DUMP , 
				   "*** unexpected message ***");
		   receive_message(rcv_msg) ; 
	       } 
	       if ((rcv_msg->type == xact) && 
		   (rcv_msg->subtype == copier_update)) {

		  num_copier++;
                  for (index = 0 ; index < rcv_msg->c.db.num_items ; index++) {
		     cur_item_value = rcv_msg->c.db.value[index];
		     data_item_mgr (WRITE_ITEM , rcv_msg->c.db.item[index] ,
				    &cur_item_value);
		     directory_mgr (CREATE_COPY, MySID, 
				  rcv_msg->c.db.item[index] ) ; 
		     fail_locks_mgr(CLEAR_FOR_SITE, MySID, 
					rcv_msg->c.db.item[index]) ;
		  }  /* end for */
               } 
            }  /* end if */

/*---------------------------------------------------------------------------
:           If a request has been made for every new item then skip the
:           other sites
---------------------------------------------------------------------------*/

            if (new_count == num_requested)  break;
	 }  /* end for */

/*---------------------------------------------------------------------------
:        Current set of locked items processed.
---------------------------------------------------------------------------*/

         free (rcv_msg);
         free (copy_msg);
      } /* end if */
      if (new_count == 0 ) 
	 check_complete = TRUE ; 
    } /* end while */
    /* restore the new directories and fail-locks */

    bcopy(msg->c.ctl_direct, directories, sizeof(directory_type)) ; 
    bcopy(msg2->c.ctl.f_locks, fail_locks, sizeof(fail_locks_type)); 

    if (rcvd_user_xact) { 
       process_user_xact(usr_msg) ; 
       free(usr_msg) ; 
    } 
} 
