/*---------------------------------------------------------------------------
:  This file contains procedures for the output of information
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
:  Header files
---------------------------------------------------------------------------*/

#  include  <stdio.h>
#  include  <strings.h>
#  include  <sys/types.h>
#  include  <sys/time.h>
#  include  "comm.h"
#  include  "simtypes.h"
#  include  "extdata.h"

/*---------------------------------------------------------------------------
:  log_site_info outputs a time stamp and a comment along with a dump of
:  status and database information if requested
---------------------------------------------------------------------------*/

log_site_info (log_function , comment)

   int  log_function;
   char *comment;
{

/*---------------------------------------------------------------------------
:  Local variable definition
---------------------------------------------------------------------------*/

#  define  MAX_OUT  5
#  define  NUM_TIME_CHARS  24
   char *time_ptr , time_string[NUM_TIME_CHARS+1];
   int index , i , num_left , num_out , num_fail_locks ;

/*DM				ctime does not take a struct timeval..
   struct timeval tp;
   struct timezone tzp;
*/
   site_id_type site;
   char fail_locks_string[MAX_NUM_SITES+1];

/*---------------------------------------------------------------------------
:  Start of code - get the current date and time
---------------------------------------------------------------------------*/

   time_t tp = time(0);

/*DM
   gettimeofday (&tp , &tzp);
*/

   time_ptr = ctime (&tp);
   strncpy (time_string , time_ptr , NUM_TIME_CHARS);

/*---------------------------------------------------------------------------
:  Log the requested information
---------------------------------------------------------------------------*/

   switch (log_function) {
      case NO_DUMP :
	   printf ("%24.24s : %s\n" , time_string , comment);
	   break;

      case DUMP :
	   dump_db (comment);	/* dump data base (fail_locks, directories) */
	   break;

      case SUMMARY :
	   printf ("%24.24s : %s\n" , time_string , comment);
	   printf ("   session vector (site|state|session num):\n      ");
	   for (site = FIRST_SITE_ID ; site < num_sites ; site++)
	      printf ("%1d|%c|%-2.2d " , site , session_vector[site].state ,
		      session_vector[site].snum);
           printf ("\n   number of fail-locks currently set:\n");
           for (site = FIRST_SITE_ID ; site < num_sites ; site++) {
	      num_fail_locks = count_fail_locks (site);
              printf ("      site %1d has %-d\n" ,
		      site , num_fail_locks);
           }
           break;

      default : 
	   printf ("%s : undefined log function\n" , time_string);
	   break;
   }
   fflush (stdout);
   return (0);
}

/*---------------------------------------------------------------------------
:  log_message outputs the contents of a message
---------------------------------------------------------------------------*/

log_message ( message )

   msg_type *message;
{
#  define  MAX_ACTIONS_OUT  7
   int index , i , num_left , num_out;

   switch (message->type) {
      case xact :
	   if (message->subtype == user) {
              printf ("    The number of actions = %-d\n" , 
		      message->c.ux.num_actions);
              printf ("    Each action has the format:  action|item|[value]\n");

              index = 0;
              while (index < message->c.ux.num_actions) {
                 printf ("      "); 
		 num_left = message->c.ux.num_actions - index;
		 num_out = (num_left > MAX_ACTIONS_OUT ? 
			    MAX_ACTIONS_OUT : num_left);
                 for (i = index ; i < index+num_out ; i++)
		    if (message->c.ux.op[i].action == READ_ITEM)
		       printf ("%c|%2.2d|xxx  " , READ_ITEM , 
			       message->c.ux.op[i].item);
                    else
		       printf ("%c|%2.2d|%3.3d  " , WRITE_ITEM ,
			       message->c.ux.op[i].item , 
			       message->c.ux.op[i].value);

                 index += num_out;
		 printf ("\n");
              }  /* end while */
           }
           else
	      printf ("*** output not supported ***\n");
           break;

      default :
	   printf ("*** output not supported ***\n");
           break;
	      
   }  /* end switch */

   return (0);
}
