/*---------------------------------------------------------------------------
:  This procedure processes a managing message type for a database site
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
:  Header files
---------------------------------------------------------------------------*/

#  include  <stdio.h>
#  include  "comm.h"
#  include  "simtypes.h"
#  include  "extdata.h"
#  define YES 1
#  define NO  0

/*---------------------------------------------------------------------------
:  Procedure entry point
---------------------------------------------------------------------------*/

process_managing_msg ( message )
  
   msg_type *message;
{

/*---------------------------------------------------------------------------
:  Local variable definition
---------------------------------------------------------------------------*/

   int log_site_info();
   site_state_type set_site_state(), get_site_state();

/*---------------------------------------------------------------------------
:  Start of code - process according to the message subtype
---------------------------------------------------------------------------*/

   switch (message->subtype) {
      case stop :
	   set_site_state (MySID , SITE_TERMINATING);
	   log_site_info (DUMP , "dump at stopping point");
	   break;

      case revive :
	   process_managing_revive( message );
	   break;

      case die :
	   failure_schedule = message->c.subfunction;
	   if (failure_schedule == FAIL_IMMEDIATELY) {
	      set_site_state (MySID , SITE_DOWN);
	      log_site_info (DUMP , "dump at failure point");
	   }
	   break;

      case dump :
	   log_site_info (DUMP , "dump of current information");
	   break;

      case failed :
	   log_site_info (NO_DUMP , "*** unsolicited1 message ***");
	   break;

      case allow_recovery :
	   if ( get_site_state(MySID) == SITE_UP) {
	   	send_recovery_response(message->c.ctl.subj_site); 
	   }
	   else log_site_info(NO_DUMP,"recd allow recovery and down");
	   break;

      default :
	   log_site_info (NO_DUMP , "*** unknown control msg subtype ***");
	   break;

   }  /* end switch */

   return (0);
}

/*------------------------------------------------------------------------
 * process_managing_revive
 *------------------------------------------------------------------------
 */
 process_managing_revive( message )
	msg_type	*message;
{
	site_state_type	get_site_state(), state;
	site_state_type	old_state;
	msg_type	recv_msg;
	int		other_sites_up;
	int		recovery_response_recd;
	int		recovery_wait_recd;
	int		recovery_announce_recd;
	int		managing_failed_recd;
	int		all_waiting_but_me;
	int		highest_one_waiting;
	int		i,k;

	if ( get_site_state(MySID) != SITE_DOWN ) 
		log_site_info(NO_DUMP,"***Recd revive when not down***");
	else {
		/*if (last_update != (msg_type *)NULL)  free (last_update);*/

		bcopy(session_vector,old_session_vector,
			sizeof(session_vector_type));

		/* if a site is up now just wait for the user to tell it
		   to issue the managing.allow_recovery message */
		for (i = FIRST_SITE_ID;i<num_sites;i++) 
			if (i != MySID){ 
				send_control_status(i);
				receive_message(&recv_msg);
				if ((recv_msg.type == managing) &&
				    (recv_msg.subtype == up)) {
				       set_site_state(MySID,SITE_RECOVERY_WAIT);
				       for(k=FIRST_SITE_ID;k<num_sites;k++)
					   if (k != MySID) 
						send_recovery_announce(k);
				       return;
				}
			}

		/* if no other site is currently up then see if ok to come up */
		other_sites_up = NO; /* up when I went down */
		for (i=FIRST_SITE_ID;i<num_sites;i++) 
			if (i != MySID) {
				set_site_state(i,SITE_STATE_UNKNOWN);
				state = get_old_site_state(i);
				if (state == SITE_UP) other_sites_up = YES;
			}
		if (other_sites_up == NO) {
printf("no other sites up\n");
			for(i=FIRST_SITE_ID;i<num_sites;i++) 
				if (i != MySID) {
					send_control_status(i);
					receive_message(&recv_msg);
					if ((recv_msg.type == managing) &&
					    (recv_msg.subtype == failed)) 
					     set_site_state(i,SITE_DOWN);
					else if ((recv_msg.type == control) &&
					  (recv_msg.subtype==recovery_announce))
				     	set_site_state(i,SITE_RECOVERY_WAIT);
					else log_site_info(DUMP,
					     "*Last site down but ERROR ***");
				}
			set_site_state(MySID,SITE_UP);
			for (i=FIRST_SITE_ID;i<num_sites; i++)
				if (get_site_state(i) == SITE_RECOVERY_WAIT)
					send_recovery_response(i);
		}

		else  { /* not the last site down */
printf("yes other sites up\n");
			recovery_response_recd = NO;
			recovery_wait_recd     = NO;
			recovery_announce_recd = NO;
			managing_failed_recd   = NO;
			for(i=FIRST_SITE_ID;i<num_sites;i++) {
		   	   if (i == MySID) continue ;
			   send_recovery_announce(i);
			   if ( ! recovery_response_recd ) {
				receive_message(&recv_msg);
				if ((recv_msg.type == control) &&
			    	    (recv_msg.subtype==recovery_response) &&
				    (recovery_response_recd == NO)) {
				       /* this is done just because process_
				          recovery_response expects it */
				       set_site_state(MySID,SITE_RECOVERY_WAIT);
				       process_recovery_response(&recv_msg);
				       recovery_response_recd = YES;
				       break;
				}
				else if ((recv_msg.type == control) &&
				         (recv_msg.subtype == recovery_wait)) {
					recovery_wait_recd = YES;
				  	if (recovery_response_recd == NO) 
				  	set_site_state(MySID, 
						SITE_RECOVERY_WAIT);
				}
				else if ((recv_msg.type == control) &&
				     (recv_msg.subtype == recovery_announce)) {
					recovery_announce_recd = YES;
					if (recovery_response_recd == NO) 
				  	   set_site_state(i,SITE_RECOVERY_WAIT);
				}
				else if ((recv_msg.type == managing) &&
				         (recv_msg.subtype == failed)) {
					managing_failed_recd == YES;
					if (recovery_response_recd == NO)
						set_site_state(i,SITE_DOWN);
				}
			   }
			} /* end for */

			if (recovery_response_recd)  {
/*-----------------------------------------------------------------------------
 * Other sites must be up 
 *-----------------------------------------------------------------------------
 */
				if (recovery_wait_recd)
					log_site_info(NO_DUMP,
						"Recovered but some sites bad");
				else	log_site_info(DUMP,
						"I successfully recovered");
			}
			else if (recovery_wait_recd) {
/*-----------------------------------------------------------------------------
 * Then I must wait for the last site that went down to come up
 *-----------------------------------------------------------------------------
 */
				if (recovery_announce_recd)
					log_site_info(NO_DUMP,
						"Wait but some site bad");
				else	log_site_info(NO_DUMP,
						"I'm successfully waiting");
			}
			else {
			
printf("figure out if all sites down at same time\n");
/*-----------------------------------------------------------------------------
 * Determine if all sites that went down at the same time are trying to 
 * recover. If so, if I am the highest ranked then send the recovery_response.
 *----------------------------------------------------------------------------- 
 */
				all_waiting_but_me = YES;
				highest_one_waiting = -1;
				for (i=FIRST_SITE_ID;i<num_sites;i++) {
					if(i != MySID) {
						state = get_site_state(i);
						old_state=get_old_site_state(i);
						if (old_state == SITE_UP) {
						 if (state==SITE_RECOVERY_WAIT)
							highest_one_waiting = i;
						 else 
							all_waiting_but_me = NO;
						}
						     
					}
				} 
printf("all waiting %d highest one %d\n", all_waiting_but_me,
					  highest_one_waiting);
				if (all_waiting_but_me) {
printf("all other site are waiting\n");
				    if (highest_one_waiting < MySID) {
printf("I am highest one coming up\n");
					for(i=FIRST_SITE_ID;i<num_sites;i++) 
					    if(get_site_state(i) == 
					        SITE_RECOVERY_WAIT)
					  	  set_site_state(i,SITE_UP);
					set_site_state(MySID,SITE_UP);
					for(i=FIRST_SITE_ID;i<num_sites;i++)
					        if (i != MySID)
						  send_recovery_response(i);
				    }
				    else log_site_info(NO_DUMP,
					"*Highest site didn't send rec_resp**");
				}
				else {
			      	   set_site_state(MySID,SITE_RECOVERY_WAIT);
printf("I guess I'll wait for other sites\n");	
				}
			} /* end else */ 
		} /* end else */
	} /* end else */
} 

